const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

const actions = new LuaMap<string, Action>()

let isInitialized = false
let isMouseoverSettingEnabled = false

interface ActionLoadConditions {
  class?: number
  spec?: number
  custom?: () => boolean
}

/**
 * Evaluates the actions load conditions to tell the ActionManager
 * wether or not this action should be loaded.
 * @param loadConditions ActionLoadConditions
 * @returns boolean
 */
function GetIsActionLoaded(this: void, loadConditions: ActionLoadConditions) {
  let matchesClass = true
  let matchesSpec = true
  let matchesCustom = true
  if (loadConditions.class !== undefined) {
    const [, , classId] = UnitClass('player')
    matchesClass = loadConditions.class === classId
  }
  if (loadConditions?.spec !== undefined) {
    matchesSpec = loadConditions.spec === GetSpecialization()
  }
  if (loadConditions.custom !== undefined) {
    matchesCustom = loadConditions.custom()
  }

  return matchesClass === true && matchesSpec === true && matchesCustom === true
}

/**
 * Returns the actions map in array format, sorted by priority.
 */
function GetSortedActions(this: void) {
  const actionsArray: Action[] = []

  for (const [, action] of actions) {
    actionsArray.push(action)
  }

  return actionsArray.sort((a, b) => {
    return a.priority - b.priority
  })
}

/**
 * Adds an action to the manager to call inside its ticker if loaded.
 * @param name A unique name to identify the action.
 * @param action The function to call for the action.
 * @param options Options for the action to add.
 */
function Add(
  this: void,
  name: string,
  options: {
    run: RunAction
    priority?: number
    loadConditions: ActionLoadConditions
  }
) {
  if (actions.has(name)) {
    throw Mekanome.Error(`Tried to add action ${name} but it already exists.`)
  }

  const loaded =
    options.loadConditions !== undefined
      ? GetIsActionLoaded(options.loadConditions)
      : false

  let priority = options.priority

  if (priority === undefined) {
    const sortedActions = GetSortedActions()

    if (sortedActions.length === 0) {
      priority = 1
    } else {
      const lastAction = sortedActions[sortedActions.length - 1]
      priority = lastAction.priority + 1
    }
  }

  actions.set(name, {
    priority,
    run: options.run,
    loadConditions: options.loadConditions,
    loaded,
    name
  })

  if (loaded) {
    Mekanome.Print({ text: name, prefix: 'Module Loaded', level: 'info' })
  }
}

/**
 * Removes the action from the list.
 * @param name The name of the action to remove.
 */
function Remove(this: void, name: string) {
  if (actions.has(name)) {
    actions.delete(name)
  }
}

/**
 * Calls the list of loaded actions, in order.
 * If an action returns a binding, that binding is run and the loop is terminted
 * for a tick.
 */
function CallActions(this: void) {
  const actionList = GetSortedActions()

  for (const action of actionList) {
    if (action && action.loaded) {
      // If an action returns a castable, dont call any other actions this tick.
      const castable = action.run()

      if (castable !== undefined) {
        Mekanome.Cast(castable)
        break
      }
    }
  }
}

/**
 * In order to cast on targets that arent the active target, we need
 * to leverage mouse over casting. If the setting isnt enabled, Mekanome
 * will not be able to cast on mobs that aren't the current target.
 */
function CheckSettings(this: void) {
  const mouseoverSetting = C_CVar.GetCVar('enableMouseoverCast')
  if (mouseoverSetting !== undefined && mouseoverSetting !== '0') {
    isMouseoverSettingEnabled = true
  } else {
    isMouseoverSettingEnabled = false
    throw Mekanome.Error('Please enable mouseover casting in the settings menu.')
  }
}

/**
 * Initializes the action manager.
 */
function Init(this: void) {
  if (isInitialized === false) {
    // Check for the right settings.
    CheckSettings()

    // Sets a ticker to call the list of actions.
    C_Timer.NewTicker(
      0.15,
      () => {
        if (Mekanome.enabled && isMouseoverSettingEnabled) {
          // Refresh the mob & traits managers every tick.
          Mekanome.MobManager.Refresh()
          Mekanome.PlayerStateManager.Refresh()
          CallActions()
        }
      },
      undefined
    )

    // Listens to the player spec change event, as that might change the load
    // status of actions in the list.
    Mekanome.EventManager.RegisterEventHandler<'PLAYER_SPECIALIZATION_CHANGED'>(
      'PLAYER_SPECIALIZATION_CHANGED',
      () => {
        for (const [name, action] of actions) {
          if (action.loadConditions) {
            const loaded = GetIsActionLoaded(action.loadConditions)
            if (loaded !== action.loaded) {
              actions.set(name, { ...action, loaded })
              Mekanome.Print({
                text: name,
                level: 'info',
                prefix: loaded ? 'Loaded Module' : 'Unloaded Module'
              })
            }
          }
        }
      },
      'am_spec_change'
    )

    // Listens to CVAR updates to make sure that mouseover casting isnt
    // disabled while Mekanome is running.
    Mekanome.EventManager.RegisterEventHandler<'CVAR_UPDATE'>(
      'CVAR_UPDATE',
      ([eventName]) => {
        if (eventName === 'enableMouseoverCast') {
          CheckSettings()
        }
      },
      'am_cvar_update'
    )

    isInitialized = true
  }
}

const _ActionManager = {
  Add,
  Remove,
  Init
}

declare global {
  type ActionManager = typeof _ActionManager
  type RunAction = (this: void) => Castable | undefined
  interface Action {
    priority: number
    name: string
    loaded: boolean
    loadConditions?: ActionLoadConditions
    run: RunAction
  }
}

Mekanome.ActionManager = _ActionManager

export default {}
