const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

const ActionBarButtonNames = [
  'ActionButton',
  'MultiBarBottomLeftButton',
  'MultiBarBottomRightButton',
  'MultiBarLeftButton',
  'MultiBarRightButton',
  'MultiBar5Button',
  'MultiBar6Button',
  'MultiBar7Button'
]

const BarMap = {
  ActionButton: 'ACTIONBUTTON',
  MultiBarBottomLeftButton: 'MULTIACTIONBAR1BUTTON',
  MultiBarBottomRightButton: 'MULTIACTIONBAR2BUTTON',
  MultiBarRightButton: 'MULTIACTIONBAR3BUTTON',
  MultiBarLeftButton: 'MULTIACTIONBAR4BUTTON',
  MultiBar5Button: 'MULTIACTIONBAR5BUTTON',
  MultiBar6Button: 'MULTIACTIONBAR6BUTTON'
}

interface Binding {
  binding: ActionButtons
  id: number | string
  type: 'item' | 'spell'
}

const bindings = new Map<number, Binding>()

let isInitialized = false

/**
 * Loops through all of the WoW action bar buttons, checks to see if there is a spell
 * or item on them, then stores that relationship for later use.
 */
function CheckBindings() {
  bindings.clear()
  for (const bar of ActionBarButtonNames) {
    for (const i of $range(1, 12)) {
      const actionButton = `${bar}${i}`

      // @ts-ignore
      const button = _G[actionButton]

      if (
        button !== undefined &&
        typeof button === 'object' &&
        button.action !== undefined
      ) {
        const [type, id] = GetActionInfo(button.action)

        if (type === 'spell' || type === 'item') {
          // @ts-ignore
          const barName = BarMap[bar]

          const binding = `${barName}${i}` as ActionButtons

          bindings.set(button.action, {
            binding,
            id,
            type
          })
        }
      }
    }
  }

  Mekanome.Print({
    text: `${bindings.size}`,
    prefix: 'Bindings Update',
    level: 'warn'
  })
}

/**
 * Returns the ActionButtons binding for a given spell.
 * @param spell
 * @returns ActionButtons
 */
function GetBindingForSpell(this: void, spell: Spell) {
  let binding: ActionButtons | undefined = undefined
  for (const [, bindingInfo] of bindings) {
    if (bindingInfo.id === spell.id) {
      binding = bindingInfo.binding
      break
    }
  }
  return binding
}

function CallBinding(this: void, binding: ActionButtons) {
  RunBinding(binding)
  RunBinding(binding, 'up')
}

/**
 * Initializes the bindings manager.
 */
function Init(this: void) {
  if (isInitialized === false) {
    CheckBindings()

    // Listen for the evet fired when a spell is added or removed from the action
    // bar so that we can update our bindings.
    Mekanome.EventManager.RegisterEventHandler<'ACTIONBAR_SLOT_CHANGED'>(
      'ACTIONBAR_SLOT_CHANGED',
      (args) => {
        const slot = args[0]
        const binding = bindings.get(slot)
        const [type, id] = GetActionInfo(slot)

        // This event gets fired for a couple of different reasons.
        // 1. A spell is added or removed from the action bar.
        // 2. A considtional macro on the action bar is update (e.g a mouseover macro when the player hovers something)
        // 3. A cast starts and ends.
        // We only care about case 1, so we only update bindings if we either dont have a binding saved
        // for this slot, or the binding is changed.
        if (binding !== undefined) {
          if (type !== binding.type || id !== binding.id) {
            CheckBindings()
          }
        } else {
          if (type === 'spell' || type === 'item') {
            CheckBindings()
          }
        }
      },
      'bm_action_bar_slot_changed'
    )

    // When you mount a dragon riding mount, and your spells get replaced with dragon
    // riding abilities this is called so we update the bindings.
    Mekanome.EventManager.RegisterEventHandler<'UPDATE_BONUS_ACTIONBAR'>(
      'UPDATE_BONUS_ACTIONBAR',
      () => {
        CheckBindings()
      },
      'bm_player_mount_display'
    )

    // Listens to the player spec change event, as that might change the load
    // status of actions in the list.
    Mekanome.EventManager.RegisterEventHandler<'PLAYER_SPECIALIZATION_CHANGED'>(
      'PLAYER_SPECIALIZATION_CHANGED',
      () => {
        CheckBindings()
      },
      'bm_spec_change'
    )

    isInitialized = true
  }
}

const _BindingsManager = {
  Init,
  GetBindingForSpell,
  CallBinding
}

declare global {
  type BindingsManager = typeof _BindingsManager
}

Mekanome.BindingsManager = _BindingsManager

export default {}
