const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

type EventCallback<T extends WoWEventKey> = (args: WowEventPayload[T]) => void

const events = new LuaMap<WowEvent, LuaMap<string, (args: any) => void>>()

const frame = CreateFrame<'Frame'>('Frame')

frame.SetScript('OnEvent', (f, event, ...args) => {
  const callbacks = events.get(event)
  if (callbacks) {
    for (const [, cb] of callbacks) {
      cb(args)
    }
  }
})

/**
 * Registers an event to listen to.
 * @param event The name of the wow event.
 * @param callback The function to call when the event triggers.
 * @param id The ID specific to this callback, so that multiple handlers can be added per event.
 */
function RegisterEventHandler<T extends WoWEventKey>(
  this: void,
  event: WowEvent,
  callback: EventCallback<T>,
  id: string
) {
  const existingHandlers = events.get(event)

  if (existingHandlers) {
    if (existingHandlers.has(id)) {
      return
    }
    existingHandlers.set(id, callback)
    events.set(event, existingHandlers)
  } else {
    frame.RegisterEvent(event)
    const handlers = new LuaMap<string, EventCallback<T>>()
    handlers.set(id, callback)
    events.set(event, handlers)
    Mekanome.Print({ text: event, prefix: 'Registered', level: 'info' })
  }
}

/**
 * Unregisters a Wow event handler.
 * @param event The event to unregister from.
 * @param id The ID of the handler to unregister.
 */
function UnregisterEventHandler(this: void, event: WowEvent, id: string) {
  const handlers = events.get(event)

  if (handlers && handlers.has(id)) {
    handlers.delete(id)
    Mekanome.Print({
      text: `Event: ${event} | ID: ${id}`,
      prefix: 'Handler',
      level: 'error'
    })
  }

  if (handlers && handlers.isEmpty() === true) {
    events.delete(event)
    frame.UnregisterEvent(event)
    Mekanome.Print({
      text: event,
      prefix: 'Event',
      level: 'error'
    })
  }
}

const _EventManager = {
  RegisterEventHandler,
  UnregisterEventHandler
}

declare global {
  type EventManager = typeof _EventManager
}

Mekanome.EventManager = _EventManager

export default {}
