const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

let isInitialized = false

// We dont really care about critters, non combat pets, and gas clouds
const CREATURE_TYPES_TO_IGNORE: UnitCreatureType[] = [8, 12, 13, 14]

// Dummys dont really place nice with the combat checking APIs, so we
// just consider th
const DUMMYS = [
  // City (SW, Orgrimar, etc)
  31146, 31144, 32666, 32667, 46647, 114832, 153285, 153292,
  // DragonFlight Valdrakken
  198594, 194648, 189632, 194643, 194644, 197833, 189617, 194649,
  // DragonFlight Iskaara
  193563,
  // Other
  65310, 66374, 196394, 196406, 199057
]

const mobs = new LuaMap<string, Mob>()

// ------------
// INTERNAL
// ------------

/**
 * Creates a "Mob" from a WowGameObject
 * @param mobObject A WowGameObject corresponding to the unit
 * @returns mob data
 */
function CreateMobFromObject(this: void, mobObject: WowGameObject): Mob {
  let type: Mob['type'] = 'unknown'

  const id = mobObject.id()
  const unit = mobObject.unit()

  if (
    UnitIsFriend('player', unit as UnitToken) ||
    UnitInParty(unit as UnitToken, undefined) ||
    UnitInRaid(unit as UnitToken, undefined)
  ) {
    type = 'friend'
  }

  if (
    UnitCanAttack('player', unit as UnitToken) ||
    (typeof id === 'number' && DUMMYS.includes(id))
  ) {
    type = 'enemy'
  }

  return {
    type,
    token: mobObject.unit(),
    guid: mobObject.guid(),
    id: mobObject.id(),
    auras: Mekanome.AuraManager.New(mobObject.unit())
  }
}

function GetMobFromObject(this: void, mobObject: WowGameObject): Mob | undefined {
  return mobs.get(mobObject.guid())
}

/**
 * Updates the time in which a mob was last in combat.
 * @param guid The GUID of the mob to update.
 */
function UpdateCombatTime(this: void, guid: string) {
  const mob = GetMob({ guid })

  if (mob && mob.type === 'enemy') {
    mob.lastCombatTime = GetTime()
    mobs.set(guid, mob)
  }
}

// ------------
// EXTERNAL
// ------------

/**
 * Returns a mob from the mob manager, if it exists
 */
function GetMob(
  this: void,
  { token, guid }: { token?: string; guid?: string }
): Mob | undefined {
  let id: string | undefined = undefined

  if (guid) {
    id = guid
  } else if (token) {
    id = token === 'none' ? token : ObjectGUID(token)
  }

  if (id === undefined) return undefined

  if (mobs.has(id) === false) {
    const gameObject = (Object as typeof TObject)(id)

    if (gameObject !== false) {
      mobs.set(id, CreateMobFromObject(gameObject))
    }
  }

  return mobs.get(id)
}

/**
 * Refreshes the list of mobs.
 */
function Refresh(this: void) {
  const touchedMobs: string[] = []
  const unitObjects = Objects(5)
  const playerObjects = Objects(6)
  const activePlayerObject = Objects(7)

  const objects = unitObjects.concat(playerObjects).concat(activePlayerObject)

  for (const gameObject of objects) {
    const objectType = ObjectCreatureType(gameObject)

    if (CREATURE_TYPES_TO_IGNORE.includes(objectType) === false) {
      const existingMob = GetMobFromObject(gameObject)
      const mob = CreateMobFromObject(gameObject)

      touchedMobs.push(mob.guid)

      if (existingMob === undefined) {
        mobs.set(mob.guid, mob)
      } else {
        const isStale = existingMob.id !== mob.id || existingMob.token !== mob.token

        if (isStale) {
          mobs.set(existingMob.guid, {
            ...existingMob,
            id: mob.id,
            token: mob.token
          })
        }
      }
    }
  }

  for (const [guid, _] of mobs) {
    if (touchedMobs.includes(guid) === false) {
      mobs.delete(guid)
    }
  }
}

/**
 * Iterates over the mob list, and calls a function for each mob.
 * @param callback Function called for each mob in the list.
 * @param type The type of mobs to iterate over. "enemies_active" iterates over enemies that are in combat.
 */
function Iterate(
  this: void,
  callback: (mob: Mob) => void,
  type?: Mob['type'] | 'enemies_active'
) {
  for (const [_, mob] of mobs) {
    if (type !== undefined) {
      if (type === 'enemies_active') {
        const combatOdds = Mekanome.MobUtils.InCombatOdds(mob)
        if (combatOdds !== undefined && combatOdds < 0.4) {
          callback(mob)
        }
      } else if (mob.type === type) {
        callback(mob)
      }
    } else {
      callback(mob)
    }
  }
}

/**
 * Initializes the mob manager.
 */
function Init(this: void) {
  if (isInitialized === false) {
    // Update mobs combat times when combat log events come through for them.
    // TODO: Only do this if the source is the player, or someone in the players party.
    Mekanome.EventManager.RegisterEventHandler<'COMBAT_LOG_EVENT_UNFILTERED'>(
      'COMBAT_LOG_EVENT_UNFILTERED',
      () => {
        if (Mekanome.enabled) {
          const [, , , source, , , , dest] = CombatLogGetCurrentEventInfo()

          UpdateCombatTime(source)
          UpdateCombatTime(dest)
        }
      },
      'mm_combat_log_event_unfiltered'
    )
    // Update mobs aura info when this event comes through and they're being managed.
    Mekanome.EventManager.RegisterEventHandler<'UNIT_AURA'>(
      'UNIT_AURA',
      ([token, auraInfo]) => {
        const mob = GetMob({ token })
        if (mob) {
          const updatedAuras = Mekanome.AuraManager.GetUpdatedAuras(auraInfo, mob)
          mob.auras = updatedAuras
          mobs.set(mob.guid, mob)
        }
      },
      'mm_unit_aura'
    )

    isInitialized = true
  }
}

const _MobManager = {
  Iterate,
  Refresh,
  GetMob,
  Init
}

declare global {
  type Mob = {
    type: 'friend' | 'enemy' | 'unknown'
    lastCombatTime?: number
    token: string
    guid: string
    id: number | boolean
    auras: Auras
  }
  type MobManager = typeof _MobManager
}

Mekanome.MobManager = _MobManager

export default {}
