declare global {
  interface WowGameObject {
    /**  Returns the Tinkr object managers reference to the object. */
    unit: () => string
    /**  Returns the objects full WoW GUID. */
    guid: () => string
    /**  Returns the units ID (can also be found in the GUID). Players dont have an ID. */
    id: () => number | boolean
  }

  function DirectoryExists(this: void, path: string): boolean
  function ListFiles(this: void, directory: string): string[]
  function ListDirectories(this: void, path: string): string[]
  function SetMouseover(this: void, mob: WowGameObject | undefined): true | false

  interface Missile {
    // Current X
    cx: number
    // Current Y
    cy: number
    // Current Z
    cz: number
    // Hit X
    hx: number
    // Hit Y
    hy: number
    // Hit Z
    hz: number
    // Init X
    ix: number
    // Init Y
    iy: number
    // Init Z
    iz: number
    // Model X
    mx: number
    // Model Y
    my: number
    // Model Z
    mz: number
    source?: WowGameObject
    target?: WowGameObject
    spellId?: number
    spellVisualId?: number
  }

  type ObjectType =
    // Object
    | 0
    // Item
    | 1
    // Container
    | 2
    // AzeriteEmpoweredItem
    | 3
    // AzeriteItem
    | 4
    // Unit
    | 5
    // Player
    | 6
    // ActivePlayer
    | 7
    // GameObject
    | 8
    // DynamicObject
    | 9
    // Corpse
    | 10
    // AreaTrigger
    | 11
    // SceneObject
    | 12
    // Conversation
    | 13
    // AiGroup
    | 14
    // Scenario
    | 15
    // Loot
    | 16
    // Invalid
    | 17

  type UnitCreatureType =
    // Beast
    | 1
    // Dragonkin
    | 2
    // Demon
    | 3
    // Elemental
    | 4
    // Giant
    | 5
    // Undead
    | 6
    // Humanoid
    | 7
    // Critter
    | 8
    // Mechanical
    | 9
    // NotSpecified
    | 10
    // Totem
    | 11
    // NonCombatPet
    | 12
    // GasCloud
    | 13
    // WildPet
    | 14
    // Aberration
    | 15

  /**
   * Returns a list of missiles
   */
  function Missiles(this: void): LuaPairsIterable<number, Missile>

  /**
   *  Returns whether or not a path exists.
   * */
  function DirectoryExists(this: void, path: string): boolean
  /**
   *  Returns an iterator of all files in the directory
   * */
  function ListFiles(this: void, path: string): LuaIterable<string>
  /**
   *  Logs to the tinkr console
   * */
  function Log(this: void, toLog: string): void

  /**
   * Will return a tokens GUID
   */
  function ObjectGUID(this: void, token: string | WowGameObject): string

  /**
   * Will return a tokens GUID
   */
  function TObject(this: void, token: string): WowGameObject | false

  /**
   * Returns a list of WowGameObjects, optionally filtered by type
   */
  function Objects(this: void, type?: ObjectType): WowGameObject[]

  /**
   * Returns a gameobject's creature type
   */
  function ObjectCreatureType(this: void, gameObject: WowGameObject): UnitCreatureType

  /**
   * Returns an objects spec number
   */
  function ObjectSpecializationID(this: void, gameObject: WowGameObject): number

  /**
   * Returns an objects height
   */
  function ObjectHeight(this: void, token: string | WowGameObject): number

  /**
   * Returns a units vector3 position
   */
  function ObjectPosition(
    this: void,
    object: WowGameObject | string
  ): LuaMultiReturn<[number, number, number]>

  /**
   * Returns the distance between two coordinates
   */
  function FastDistance(
    this: void,
    x: number,
    y: number,
    z: number,
    x2: number,
    y2: number,
    z2: number
  ): number

  /**
   * Returns the coords of the specified attachment position
   */
  function GetUnitAttachmentPosition(
    this: void,
    token: string | WowGameObject,
    position: number
  ): LuaMultiReturn<[number, number, number]>

  /**
   * Traces a line between two coords, and returns a position if there was a collision while tracing
   */
  function TraceLine(
    this: void,
    x1: number,
    y1: number,
    z1: number,
    x2: number,
    y2: number,
    z2: number,
    flag: number
  ): LuaMultiReturn<[number, number, number]>

  /**
   * Returns the object combat reach
   */
  function ObjectCombatReach(this: void, token: string): number | false

  /**
   * Returns the game object of the objects current cast target
   */
  function ObjectCastingTarget(
    this: void,
    token: string | WowGameObject
  ): WowGameObject | false

  /**
   * Returns the objects movement bit. See https://docs.tinkr.site/Lua/Objects/ObjectMovementFlag/
   */
  function ObjectMovementFlag(this: void, token: string | WowGameObject): number

  /**
   * Returns the objects rotation
   */
  function ObjectRotation(this: void, token: string | WowGameObject): number | false

  /**
   * Returns wether or not the unit is sitting
   */
  function UnitIsSitting(this: void, token: string | WowGameObject): boolean

  /**
   * Returns the objects model id
   */
  function ObjectModelId(this: void, token: string | WowGameObject): number | false

  /**
   * Returns the objects object manager ID
   */
  function ObjectId(this: void, token: string | WowGameObject): number | false

  /**
   * Returns whether or not the unit is mounted
   */
  function UnitIsMounted(this: void, token: string | WowGameObject): boolean

  /**
   * Returns the target object of the object
   */
  function ObjectTarget(
    this: void,
    token: string | WowGameObject
  ): WowGameObject | undefined

  /**
   * Returns whether or not a spell is pending (green cursor)
   */
  function IsSpellPending(this: void): number

  /**
   * Clicks on the specified position.
   */
  function Click(this: void, x: number, y: number, z: number): void

  interface tinkr {
    Util: {
      Evaluator: {
        LoadString: (toLoad: string, args: any[], filePath: string) => any
      }
      Commands: {
        New(prefix: string): {
          Register: (prefix: string, cb: () => void, helpMessage?: string) => void
        }
      }
      File: {
        Read: (filePath: string) => string
        Run: (filePath: string, args?: any[]) => any
        LoadPackage: (filePath: string) => void
      }
    }
    Common: {
      GetAnglesBetweenPositions(
        x1: number,
        y1: number,
        z1: number,
        x2: number,
        y2: number,
        z2: number
      ): number
    }
  }

  // Overrides of WoW Docs
  namespace C_Timer {
    function After(this: void, seconds: number, callback: () => void): void

    function NewTicker(
      this: void,
      seconds: number,
      callback: () => void,
      iterations: number | undefined
    ): { IsCancelled(): boolean; Cancel(): void }

    function NewTimer(
      this: void,
      seconds: number,
      callback: () => void
    ): { IsCancelled(): boolean; Cancel(): void }
  }

  const SlashCmdList: { [key: string]: (this: void, message: string) => void }
}

export {}
