const [, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

/**
 * Check whether or not the mob is a valid wow mob.
 */
function IsValid(this: void, mob: Mob) {
  return (
    mob.token !== 'none' &&
    mob.token !== undefined &&
    Mekanome.GetObject(mob.token) !== false
  )
}

/**
 * Get whether or not the mob is alive.
 */
function IsAlive(this: void, mob: Mob) {
  return UnitIsDeadOrGhost(mob.token as UnitToken) === false
}

/**
 * Get whether or not the mob is a pet.
 */
function IsPet(this: void, mob: Mob): boolean {
  return UnitIsUnit(mob.token, 'pet')
}

/**
 * Get whether or not the player can attack this mob.
 */
function CanAttack(this: void, mob: Mob): boolean {
  const token = mob.token as UnitToken
  const isFriend = UnitIsFriend('player', token)

  if (isFriend) return false

  const isAttackable = UnitCanAttack('player', token)

  if (isAttackable === false) return false

  const reaction = UnitReaction('player', token)

  if (reaction === undefined) {
    return false
  } else {
    const isPositiveReaction = reaction >= 5
    if (isPositiveReaction) return false
  }

  return true
}

/**
 * Get the name of the mob.
 */
function GetName(this: void, mob: Mob) {
  const [unitName] = UnitName(mob.token)
  return unitName
}

/**
 * Get the health of the mob.
 */
function GetHealth(this: void, mob: Mob) {
  const token = mob.token as UnitToken

  const absorbAmount = UnitGetTotalHealAbsorbs(token)
  const health = UnitHealth(token, false) - absorbAmount

  return health
}

/**
 * Get the mobs current health expressed as a percentage of their max health.
 */
function GetHealthPercent(this: void, mob: Mob) {
  const health = GetHealth(mob)
  const maxHealth = UnitHealthMax(mob.token as UnitToken)

  return (health / maxHealth) * 100
}

/**
 * Get the mobs default powertype, eg 1 for Rage.
 */
function GetPowerType(this: void, mob: Mob) {
  const [powerType] = UnitPowerType(mob.token as UnitToken, 0)
  return powerType
}

/**
 * Get the mobs maximum power of their default power type, or a specified one.
 */
function GetMaxPower(this: void, mob: Mob, powerType?: PowerType) {
  const maxPower = UnitPowerMax(
    mob.token as UnitToken,
    powerType || GetPowerType(mob),
    false
  )
  return maxPower
}

/**
 * Get the current value for the mobs default powertype, or a specified one.
 */
function GetCurrentPower(this: void, mob: Mob, powerType?: PowerType) {
  const power = UnitPower(mob.token as UnitToken, powerType || GetPowerType(mob), false)
  return power
}

/**
 * Get the current value for the mobs default powertype, or a specified one, expressed as a percentage of the max for that powertype.
 */
function GetPowerPercent(this: void, mob: Mob, powerType: PowerType) {
  const power = GetCurrentPower(mob, powerType)
  const maxPower = GetMaxPower(mob, powerType)

  return (power / maxPower) * 100
}

type ROLE = 'TANK' | 'HEALER' | 'DAMAGER'

/**
 * Gets the group role of the mob if it exists.
 */
function GetRole(this: void, mob: Mob): ROLE | undefined {
  return UnitGroupRolesAssigned(mob.token as UnitToken) as ROLE | undefined
}

/**
 * Check whether or not the mob is in combat according to the wow api.
 */
function IsAffectingCombat(this: void, mob: Mob) {
  return UnitAffectingCombat(mob.token as UnitToken)
}

/**
 * Returns the Vector3 coords of the mob.
 */
function GetPosition(this: void, mob: Mob): Vector3 {
  const [x, y, z] = ObjectPosition(mob.token)
  return Mekanome.VectorUtils.Create({ x, y, z })
}

/**
 * Returns distance between two mobs.
 */
function GetDistanceBetween(this: void, referenceMob: Mob, compareMob: Mob): number {
  return Mekanome.VectorUtils.Distance(GetPosition(referenceMob), GetPosition(compareMob))
}

/**
 * Get whether or not a mob can see another mob un-obstructed.
 */
function CanSee(this: void, reference: Mob, compare: Mob): boolean {
  const [ax, ay, az] = ObjectPosition(reference.token)
  const ah = ObjectHeight(reference.token)
  const [attx, atty, attz] = GetUnitAttachmentPosition(compare.token, 34)

  if (attx === undefined || ax === undefined || ah === undefined) return false

  if ((ax === 0 && ay === 0 && az === 0) || (attx === 0 && atty === 0 && attz == 0)) {
    return true
  }

  const [x, y, z] = TraceLine(ax, ay, az + ah, attx, atty, attz, 0)

  if (x !== 0 || y !== 0 || z !== 0) {
    return false
  } else {
    return true
  }
}

/**
 * Returns whether or not a mob is casting a spell
 */
function IsCasting(this: void, mob: Mob) {
  return UnitCastingInfo(mob.token as UnitToken) !== undefined
}

/**
 * Returns whether or not a mob is channeling a spell
 */
function IsChanneling(this: void, mob: Mob) {
  return UnitChannelInfo(mob.token as UnitToken) !== undefined
}

/**
 * Returns whether or not a mob is channeling or casting a spell
 */
function IsCastingOrChanneling(this: void, mob: Mob) {
  return IsChanneling(mob) || IsCasting(mob)
}

function GetEndTime(this: void, startMs: number, endMs: number, percent?: number) {
  const castLength = endMs - startMs
  const startTime = startMs / 1000
  const timeUntil = (castLength / 1000) * ((percent || 100) / 100)
  return startTime + timeUntil
}

/**
 * If a mob is casting a spell, return the time when it will end.
 * @param percent: By default the cast end time is when the cast is complete. If you want to consider
 * a spell as being done once it reaches a certain percentage threshold, pass in this value.
 */
function GetCastEndTime(this: void, mob: Mob, percent?: number) {
  const castingInfo = UnitCastingInfo(mob.token as UnitToken)

  if (castingInfo !== undefined) {
    const name = castingInfo[0]
    const startTimeMS = castingInfo[3]
    const endTimeMs = castingInfo[4]

    if (name !== undefined && startTimeMS !== undefined && endTimeMs !== undefined) {
      return GetEndTime(startTimeMS, endTimeMs, percent)
    }
  }
  return undefined
}

/**
 * If a mob is channeling a spell, return the time when it will end.
 * @param percent: By default the cast end time is when the channel ends. If you want to consider
 * a spell as being done once it reaches a certain percentage threshold, pass in this value.
 */
function GetChannelEndTime(this: void, mob: Mob, percent?: number) {
  const channelInfo = UnitChannelInfo(mob.token as UnitToken)

  if (channelInfo !== undefined) {
    const name = channelInfo[0]
    const startTimeMS = channelInfo[3]
    const endTimeMs = channelInfo[4]

    if (name !== undefined && startTimeMS !== undefined && endTimeMs !== undefined) {
      return GetEndTime(startTimeMS, endTimeMs, percent)
    }
  }
  return undefined
}

function GetCastOrChannelEndTime(
  this: void,
  mob: Mob,
  percent?: number
): number | undefined {
  return GetCastEndTime(mob, percent) || GetChannelEndTime(mob, percent)
}

/**
 * If the mob is casting or channeling a spell, this returns the target mob of said spell.
 */
function GetCastTarget(this: void, mob: Mob) {
  const isMobCasting = IsCastingOrChanneling(mob)

  if (isMobCasting) {
    const target = ObjectCastingTarget(mob.token)
    if (target !== false) {
      const targetMob = Mekanome.MobManager.GetMob({ guid: target.guid() })

      return targetMob || Mekanome.MobManager.GetMob({ token: 'none' })
    } else {
      return Mekanome.MobManager.GetMob({ token: 'none' })
    }
  }
  return undefined
}

/**
 * If the mob is casting or channeling a spell, this returns how far into the cast the mob is, as a percentage.
 */
function GetCastOrChannelPercentComplete(this: void, mob: Mob) {
  let name: string | undefined = undefined
  let startTimeMS: number | undefined = undefined
  let endTimeMs: number | undefined = undefined

  const castingInfo = UnitCastingInfo(mob.token as UnitToken)

  if (castingInfo === undefined) {
    const channelInfo = UnitChannelInfo(mob.token as UnitToken)

    if (channelInfo !== undefined) {
      name = channelInfo[0]
      startTimeMS = channelInfo[3]
      endTimeMs = channelInfo[4]
    }
  } else {
    name = castingInfo[0]
    startTimeMS = castingInfo[3]
    endTimeMs = castingInfo[4]
  }

  if (name !== undefined && startTimeMS !== undefined && endTimeMs !== undefined) {
    const start = startTimeMS / 1000
    const finish = endTimeMs / 1000
    const current = GetTime()
    return ((current - start) / (finish - start)) * 100
  }

  return undefined
}

/**
 * If the mob is casting or channeling a spell, returns whether or not that spell is interruptible.
 */
function GetIsInterruptible(this: void, mob: Mob) {
  let name: string | undefined = undefined
  let notInterruptible: boolean | undefined = undefined

  const castingInfo = UnitCastingInfo(mob.token as UnitToken)

  if (castingInfo === undefined) {
    const channelInfo = UnitChannelInfo(mob.token as UnitToken)
    if (channelInfo !== undefined) {
      name = channelInfo[0]
      notInterruptible = channelInfo[6]
    }
  } else {
    name = castingInfo[0]
    notInterruptible = castingInfo[7]
  }

  if (name !== undefined && notInterruptible !== undefined) {
    return notInterruptible === false
  }

  return false
}

/**
 * Returns whether or not a mob is interruptible at a specific percentage of their cast.
 * @param interruptPercent The percent to check against.
 * @param ignoreInterruptible By default this will return false if the active spell is not interruptible. This bypasses that check.
 */
function GetIsInterruptibleAt(
  this: void,
  mob: Mob,
  interruptPercent: number,
  ignoreInterruptible?: boolean
) {
  if (!ignoreInterruptible && GetIsInterruptible(mob) === false) {
    return false
  }

  const castPercent = GetCastOrChannelPercentComplete(mob)

  if (castPercent && castPercent >= interruptPercent) {
    return true
  }

  return false
}

/**
 * Gets whether or not the mob is moving.
 */
function GetIsMoving(this: void, mob: Mob) {
  const [currentSpeed] = GetUnitSpeed(mob.token as UnitToken)

  return currentSpeed > 0
}

/**
 * Checks if a mob is facing another using their current positions.
 */
function IsFacingMob(this: void, reference: Mob, compare: Mob) {
  const rotation = ObjectRotation(reference.token)

  const [x, y] = ObjectPosition(reference.token)
  const [x2, y2] = ObjectPosition(compare.token)

  if (!x || !x2 || !rotation) {
    return false
  }

  let angle = math.atan2(y2 - y, x2 - x) - rotation
  angle = math.deg(angle)
  angle = angle % 360

  if (angle > 180) {
    angle = angle - 360
  }

  return math.abs(angle) < 90
}

/**
 * Checks if a mob is behind another using their current positions.
 */
function IsBehindMob(this: void, reference: Mob, compare: Mob) {
  const rotation = ObjectRotation(reference.token)
  const [x, y] = ObjectPosition(reference.token)
  const [x2, y2] = ObjectPosition(compare.token)

  if (!x || !x2 || !rotation) {
    return false
  }

  let angle = math.atan2(y2 - y, x2 - x) - rotation
  angle = math.deg(angle)
  angle = angle % 360

  if (angle > 180) {
    angle = angle - 360
  }

  return math.abs(angle) > 90
}

/**
 * Gets the model ID for the mob.
 */
function GetModelId(this: void, mob: Mob) {
  return ObjectModelId(mob.token)
}

/**
 * Returns whether or not the mob is in a party with the player.
 */
function IsInParty(this: void, mob: Mob) {
  return UnitInParty(mob.token as UnitToken, undefined)
}

/**
 * Returns whether or not the mob is in a raid with the player.
 */
function IsInRaid(this: void, mob: Mob) {
  return UnitInRaid(mob.token as UnitToken, undefined) !== undefined
}

/**
 * Returns whether or not the mob is in a party or raid with the player.
 */
function IsInPartyOrRaid(this: void, mob: Mob) {
  return IsInParty(mob) || IsInRaid(mob)
}

/**
 * Returns whether or not the player is mounted, or shapeshifted into a mount form.
 */
function IsMounted(this: void, mob: Mob) {
  const isMounted = UnitIsMounted(mob.token)

  if (isMounted) return true

  const mountFormIds = [
    3, // Mount / Travel form
    27, // Swift Flight Form
    29 // Flight Form
  ]

  const shapeShiftId = GetShapeshiftFormID()

  return shapeShiftId !== undefined && mountFormIds.includes(shapeShiftId)
}

/**
 * Returns the "Combat reach" of a mob. The combat reach of a mob is a bounding radius from which distance calcs begin for the purposes of calculating combat distance.
 */
function GetCombatReach(this: void, mob: Mob) {
  return ObjectCombatReach(mob.token) || 0
}

/**
 * Returns the distance between two mobs, with combat reach accounted for.
 */
function GetCombatDistanceBetween(this: void, reference: Mob, compare: Mob) {
  GetDistanceBetween(reference, compare) - GetCombatReach(compare)
}

/**
 * Gets whether or not the mob is currently online.
 */
function GetIsOnline(this: void, mob: Mob) {
  return UnitIsConnected(mob.token as UnitToken)
}

/**
 * Get whether or not the mob is being resurrected.
 */
function HasIncomingRessurection(this: void, mob: Mob) {
  return IsAlive(mob) === false && UnitHasIncomingResurrection(mob.token)
}

/**
 * Get whether or not the mob is currently targetting something.
 */
function HasTarget(this: void, mob: Mob) {
  return ObjectTarget(mob.token) !== undefined
}

/**
 * Get the mobs current target.
 */
function GetTarget(this: void, mob: Mob): Mob | undefined {
  const objTarget = ObjectTarget(mob.token)

  if (!objTarget) return undefined

  const mobManagerMob = Mekanome.MobManager.GetMob({ token: objTarget.unit() })

  return mobManagerMob
}

/**
 * Returns whether or not a mob is within a certain distance from another mob, with combat reach accounted for.
 */
function IsWithinCombatDistance(
  this: void,
  reference: Mob,
  compare: Mob,
  distance: number
) {
  if (!IsValid(compare)) return false

  return GetDistanceBetween(reference, compare) <= distance + GetCombatReach(compare)
}

/**
 * Returns whether or not a mob is within a certain distance from another mob.
 */
function IsWithinDistance(this: void, reference: Mob, compare: Mob, distance: number) {
  return GetDistanceBetween(reference, compare) <= distance
}

/**
 * Returns the number of loss of control effects on the mob. Interrupt lockouts included.
 */
function GetLossOfControlCount(this: void, mob: Mob) {
  return C_LossOfControl.GetActiveLossOfControlDataCountByUnit(mob.token as UnitToken)
}

/**
 * Returns the mobs outgoing missiles.
 */
function GetOutgoingMissles({
  reference,
  target,
  spellId,
  spellVisualId
}: {
  reference: Mob
  target?: Mob
  spellId?: number
  spellVisualId?: number
}) {
  const missiles = Missiles()

  const results: Missile[] = []

  if (type(missiles) == 'table' && IsValid(reference)) {
    for (const [, missile] of missiles) {
      const missileSource = missile.source?.unit()

      if (missileSource) {
        if (
          UnitIsUnit(reference.token, missileSource) &&
          (!target || UnitIsUnit(missile.target?.unit() || 'none', target.token)) &&
          (!spellId || (!!spellId && spellId === missile.spellId)) &&
          (!spellVisualId || (!!spellVisualId && spellVisualId == missile.spellVisualId))
        ) {
          results.push(missile)
        }
      }
    }
  }
  return results
}

/**
 * Returns the mobs incoming missiles.
 */
function GetIncomingMissiles(
  this: void,
  {
    reference,
    source,
    spellId,
    spellVisualId
  }: { reference: Mob; source?: Mob; spellId?: number; spellVisualId?: number }
) {
  const missiles = Missiles()

  const results: Missile[] = []
  if (type(missiles) == 'table' && IsValid(reference)) {
    for (const [, missile] of missiles) {
      const missileTarget = missile.target?.unit()
      if (
        missileTarget &&
        UnitIsUnit(reference.token, missileTarget) &&
        (!source || UnitIsUnit(source.token || 'none', reference.token)) &&
        (!spellId || spellId == missile.spellId) &&
        (!spellVisualId || spellVisualId == missile.spellVisualId)
      ) {
        table.insert(results, missile)
      }
    }
  }
  return results
}

/**
 * Gets the amount of time the mob has been in combat.
 */
function GetCombatTime(this: void, mob: Mob) {
  if (!mob.lastCombatTime) return undefined
  return GetTime() - mob.lastCombatTime
}

/**
 * Get combat odds (if the last combat time is less than 1 minute ago return 1 / time, else return 0)
 * the closer to 0 the more likely the unit is to be in combat (0 = 100%) 60 = 0%
 * @param mob The mob to check
 * @returns number | undefined
 */
function InCombatOdds(this: void, mob: Mob) {
  const combatTime = GetCombatTime(mob)

  if (combatTime === undefined) return undefined

  const percent = 1 - combatTime / 60

  return percent * 100
}

const _MobUtils = {
  GetLossOfControlCount,
  GetTarget,
  IsWithinCombatDistance,
  IsWithinDistance,
  GetOutgoingMissles,
  GetIncomingMissiles,
  HasTarget,
  HasIncomingRessurection,
  GetCombatDistanceBetween,
  GetCombatReach,
  IsInPartyOrRaid,
  IsMounted,
  GetModelId,
  IsInRaid,
  IsInParty,
  IsBehindMob,
  GetIsOnline,
  IsFacingMob,
  GetIsMoving,
  GetIsInterruptible,
  GetCastOrChannelPercentComplete,
  IsCasting,
  GetCastOrChannelEndTime,
  IsChanneling,
  GetIsInterruptibleAt,
  IsCastingOrChanneling,
  IsAffectingCombat,
  GetPowerType,
  CanSee,
  GetMaxPower,
  GetCurrentPower,
  GetPowerPercent,
  GetCastTarget,
  InCombatOdds,
  GetName,
  CanAttack,
  GetHealthPercent,
  IsAlive,
  GetPosition,
  GetDistanceBetween,
  IsPet,
  GetRole,
  GetCombatTime,
  GetChannelEndTime,
  GetCastEndTime
}

declare global {
  type MobUtils = typeof _MobUtils
}

Mekanome.MobUtils = _MobUtils

export default {}
