const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

/**
 * Creates a useable spell
 */
function Create(this: void, id: number, traits?: SpellTraits): Spell {
  const spellInfo = C_Spell.GetSpellInfo(id)

  if (spellInfo === undefined || spellInfo.name === undefined) {
    throw Mekanome.Error(`Unable to find info for spell ${id}`)
  }

  return {
    id,
    minRange: spellInfo.minRange,
    maxRange: spellInfo.maxRange,
    name: spellInfo.name,
    traits
  }
}

/**
 * Get cooldown information for the spell.
 */
function GetCooldownInfo(this: void, spell: Spell) {
  const [, cdDuration, enabled] = GetSpellCooldown(spell.id)
  const [, ccDuration] = GetSpellLossOfControlCooldown(spell.id)

  const { spellQueueWindow } = Mekanome.PlayerStateManager.GetPlayerState()

  const secondsUntilReady = cdDuration === 0 ? ccDuration : cdDuration
  const secondsUntilCastable = secondsUntilReady - spellQueueWindow / 1000

  return {
    secondsUntilReady,
    secondsUntilCastable: secondsUntilCastable < 0 ? 0 : secondsUntilCastable,
    isActive: enabled === 0
  }
}

/**
 * Handles undefined / funciton traits to get a more usuable set of trait values
 */
function GetSpellTraits(this: void, spell: Spell) {
  let isCastableWhileMoving = false
  let isCastableWhileCasting = false
  let isOverride = false

  if (spell.traits !== undefined) {
    if (spell.traits.isOverride !== undefined) {
      isOverride = spell.traits.isOverride
    }
    if (spell.traits.isCastableWhileCasting !== undefined) {
      if (typeof spell.traits.isCastableWhileCasting === 'function') {
        isCastableWhileCasting = spell.traits.isCastableWhileCasting()
      } else {
        isCastableWhileCasting = spell.traits.isCastableWhileCasting
      }
    }
    if (spell.traits.isCastableWhileMoving !== undefined) {
      if (typeof spell.traits.isCastableWhileMoving === 'function') {
        isCastableWhileMoving = spell.traits.isCastableWhileMoving()
      } else {
        isCastableWhileMoving = spell.traits.isCastableWhileMoving
      }
    }
  }

  return {
    isCastableWhileMoving,
    isCastableWhileCasting,
    isOverride
  }
}

interface IsCastableArgs {
  spell: Spell
  target?: Mob
  position?: Vector3
}

/**
 * Evaluates the spells traits & conditions to see if it is currently castable.
 */
function IsCastable(this: void, { spell, target, position }: IsCastableArgs): boolean {
  const player = Mekanome.MobManager.GetMob({ token: 'player' })

  const { isCastableWhileCasting, isCastableWhileMoving, isOverride } =
    GetSpellTraits(spell)

  if (player === undefined) return false

  const traits = Mekanome.PlayerStateManager.GetPlayerState()

  // Return false is the spell is not known.
  if (isOverride === true) {
    if (IsSpellKnownOrOverridesKnown(spell.id) !== true) {
      return false
    }
  } else {
    if (IsSpellKnown(spell.id) !== true) {
      return false
    }
  }

  const { isActive, secondsUntilCastable } = GetCooldownInfo(spell)

  print(`COOLDOWN: ${secondsUntilCastable}`)

  // If the spell is current active (e.g Presence of Mind was cast, but has not been consumed yet)
  // return false.
  if (isActive) return false

  // Return false if the spell is on
  if (secondsUntilCastable !== 0) {
    return false
  }

  // Return falst is the player is moving and the spell isnt marked
  // as castable while moving.
  if (traits.isMoving && isCastableWhileMoving === false) {
    return false
  }

  // Return false if the player is currently casting, and this spell
  // cant be cast while casting.
  if (traits.isCasting === true && isCastableWhileCasting === false) {
    return false
  }

  return true
}

/**
 * Get the localized name of the spell.
 */
function GetName(this: void, spell: Spell) {
  const name = C_Spell.GetSpellInfo(spell.id).name
  return name
}

/**
 * Gets the action button binding for the provided spell.
 */
function GetBindingForSpell(this: void, spell: Spell) {
  const binding = Mekanome.BindingsManager.GetBindingForSpell(spell)

  if (binding === undefined) {
    throw Mekanome.Error(`Unable to cast ${GetName(spell)} without binding.`)
  }

  return binding
}

/**
 * Generates a castable to cast the spell on the player
 */
function SelfCast(this: void, spell: Spell): Castable | undefined {
  const binding = GetBindingForSpell(spell)

  if (IsCastable({ spell }) === false) {
    return undefined
  }

  return {
    id: spell.id,
    variant: 'self',
    binding
  }
}

/**
 * Generates a castable to cast the spell at the specified target.
 */
function TargetCast(
  this: void,
  spell: Spell,
  _target: Mob | (() => Mob)
): Castable | undefined {
  const binding = GetBindingForSpell(spell)

  const target = typeof _target === 'function' ? _target() : _target

  if (IsCastable({ spell, target }) === false) {
    return undefined
  }

  return {
    id: spell.id,
    variant: 'target',
    binding,
    target
  }
}

/**
 * Generates a castable to cast the spell at the specified location
 */
function GroundCast(
  this: void,
  spell: Spell,
  _position: Vector3 | (() => Vector3)
): Castable | undefined {
  const binding = GetBindingForSpell(spell)

  const position = typeof _position === 'function' ? _position() : _position

  if (IsCastable({ spell, position }) === false) {
    return undefined
  }

  return {
    id: spell.id,
    variant: 'ground',
    binding,
    position
  }
}

const _SpellUtils = {
  Create,
  GetName,
  GetCooldownInfo,
  SelfCast,
  GroundCast,
  TargetCast
}

declare global {
  interface SpellTraits {
    isCastableWhileMoving?: boolean | (() => boolean)
    isCastableWhileCasting?: boolean | (() => boolean)
    isOverride?: boolean
  }

  interface Spell {
    id: number
    name: string
    minRange: number
    maxRange: number
    traits?: SpellTraits
  }
  type SpellUtils = typeof _SpellUtils
}

Mekanome.SpellUtils = _SpellUtils

export default {}
