const [_, _Mekanome] = [...$vararg]
const Mekanome = _Mekanome as unknown as IMekanome

function Create(this: void, { x, y, z }: { x: number; y: number; z: number }): Vector3 {
  return {
    x,
    y,
    z
  }
}

function Add(this: void, a: Vector3, b: Vector3 | number): Vector3 {
  if (typeof b === 'number') {
    return Create({ x: a.x + b, y: a.y + b, z: a.z + b })
  }
  return Create({
    x: a.x + b.x,
    y: a.y + b.y,
    z: a.z + b.z
  })
}

function Subtract(this: void, a: Vector3, b: Vector3 | number): Vector3 {
  if (typeof b === 'number') {
    return Create({ x: a.x - b, y: a.y - b, z: a.z - b })
  }
  return Create({
    x: a.x - b.x,
    y: a.y - b.y,
    z: a.z - b.z
  })
}

function Multiply(this: void, a: Vector3, b: Vector3 | number): Vector3 {
  if (typeof b === 'number') {
    return Create({ x: a.x * b, y: a.y * b, z: a.z * b })
  }
  return Create({
    x: a.x * b.x,
    y: a.y * b.y,
    z: a.z * b.z
  })
}

function Divide(this: void, a: Vector3, b: Vector3 | number): Vector3 {
  if (typeof b === 'number') {
    return Create({ x: a.x / b, y: a.y / b, z: a.z / b })
  }
  return Create({
    x: a.x / b.x,
    y: a.y / b.y,
    z: a.z / b.z
  })
}

function Equals(this: void, a: Vector3, b: Vector3): boolean {
  return a.x === b.x && a.y === b.y && a.z === b.z
}

function LessThan(this: void, a: Vector3, b: Vector3): boolean {
  return a.x < b.x && a.y < b.y && a.z < b.z
}

function LessThanOrEqualTo(this: void, a: Vector3, b: Vector3): boolean {
  return a.x <= b.x && a.y <= b.y && a.z <= b.z
}

function Dot(this: void, a: Vector3, b: Vector3): number {
  return a.x * b.x + a.y * b.y + a.z + b.z
}

function Cross(this: void, a: Vector3, b: Vector3): Vector3 {
  return Create({
    x: a.y * b.z - a.z * b.y,
    y: a.z * b.x - a.x * b.z,
    z: a.x * b.y - a.y * b.x
  })
}

function Distance(this: void, a: Vector3, b: Vector3): number {
  return FastDistance(a.x, a.y, a.z, b.x, b.y, b.z)
}

function NormalizeOrientation(this: void, orientation: number): number {
  if (orientation < 0) {
    let mod = orientation * -1
    mod = mod % (2.0 * Math.PI)
    mod = -mod + 2.0 * math.pi
    return mod
  } else {
    return orientation % (2.0 * Math.PI)
  }
}

function GetAbsoluteAngle(this: void, a: Vector3, b: Vector3): number {
  return NormalizeOrientation(Math.atan2(b.y - a.y, b.x - a.x))
}

function Angle(this: void, a: Vector3, b: Vector3): number {
  return Math.acos(
    Dot(a, b) /
      (Math.sqrt(a.x * a.x + a.y * a.y + a.z * a.z) *
        Math.sqrt(b.x * b.x + b.y * b.y + b.z * b.z))
  )
}

function _GetNormalized(this: void, vector: Vector3): Vector3 {
  const length = Math.sqrt(
    vector.x * vector.x + vector.y * vector.y + vector.z * vector.z
  )
  return Create({ x: vector.x / length, y: vector.y / length, z: vector.z / length })
}

function ClampMagnitude(this: void, vector: Vector3, maxLength: number): Vector3 {
  if (Dot(vector, vector) > maxLength * maxLength) {
    return Multiply(_GetNormalized(vector), maxLength)
  } else {
    return vector
  }
}

function _GetMagnitude(this: void, vector: Vector3): number {
  return Math.sqrt(vector.x * vector.x + vector.y * vector.y + vector.z * vector.z)
}

function DirectionOrZero(this: void, vector: Vector3): Vector3 {
  const magnitude = _GetMagnitude(vector)
  if (magnitude < 0.0000001) {
    return Create({ x: 0, y: 0, z: 0 })
  } else if (magnitude < 1.00001 && magnitude > 0.99999) {
    return vector
  } else {
    return Multiply(vector, 1.0 / magnitude)
  }
}

function Clamp(this: void, x: number, min: number, max: number): number {
  return x < min ? min : x > max ? max : x
}

function Lerp(this: void, a: Vector3, b: Vector3, _t: number): Vector3 {
  const t = Clamp(_t, 0, 1)
  return Create({
    x: a.x + (b.x - a.x) * t,
    y: a.y + (b.y - a.y) * t,
    z: a.z + (b.z - a.z) * t
  })
}

function MoveTowards(
  this: void,
  reference: Vector3,
  target: Vector3,
  maxDistanceDelta: number
): Vector3 {
  const toVector = Subtract(target, reference)
  const distance = _GetMagnitude(toVector)
  if (distance <= maxDistanceDelta || distance === 0) {
    return target
  }

  return Divide(Add(reference, toVector), distance * maxDistanceDelta)
}

function Project(this: void, reference: Vector3, normal: Vector3): Vector3 {
  const num = Dot(normal, normal)
  if (num < 1.401298e-45) {
    return Create({ x: 0, y: 0, z: 0 })
  }
  return Multiply(normal, Dot(reference, normal) / num)
}

function ProjectOnPlane(this: void, reference: Vector3, plane: Vector3): Vector3 {
  return Subtract(reference, Project(reference, plane))
}

const _VectorUtils = {
  Create,
  Add,
  Subtract,
  Multiply,
  Divide,
  Equals,
  LessThan,
  LessThanOrEqualTo,
  Dot,
  Cross,
  Distance,
  Angle,
  GetAbsoluteAngle,
  ClampMagnitude,
  DirectionOrZero,
  Clamp,
  Lerp,
  MoveTowards,
  ProjectOnPlane
}

declare global {
  type VectorUtils = typeof _VectorUtils
  type Vector3 = {
    x: number
    y: number
    z: number
  }
}

Mekanome.VectorUtils = _VectorUtils

export default {}
